// power_condBranch_LRAndR.C
// See the .h file for comments

#include "power_condBranch_LRAndR.h"
#include "power_instr.h"
#include "directToMemoryEmitter.h"
#include "funkshun.h"


condBranch_LRAndR::condBranch_LRAndR(kptr_t iLaunchSiteAddr,
				     const regset_t *iFreeIntRegsBeforeInsn,
				     const regset_t *iFreeIntRegsAfterInsn,
				     const instr_t *ioriginsn,
				     SpringBoardHeap &isbheap) : LRAndR(iLaunchSiteAddr,
								iFreeIntRegsBeforeInsn,
								iFreeIntRegsAfterInsn,
								ioriginsn, 
								isbheap)
{
   power_instr::power_cfi cfi;
   ioriginsn->getControlFlowInfo(&cfi);
   assert(cfi.fields.controlFlowInstruction && cfi.power_fields.condTrue);
   if (cfi.destination == pcRelative)
     condBranch_destination = iLaunchSiteAddr + cfi.offset_nbytes;
   else {
     assert(cfi.destination == absolute);
     condBranch_destination = cfi.offset_nbytes;
   }
       
}

unsigned condBranch_LRAndR::conservative_numInsnBytesForRAndR() const {
  return 4*(1+1+1);
  //worst case: short-range conditional jump + unconditional jump to destination + not-taken jump
}

void condBranch_LRAndR::
registerPatchAddrAndCreateLauncher(kptr_t ipatchAddr,
                                   const instr_t *whenAllDoneRestoreToThisInsn) {
   setPatchAddr(ipatchAddr);
   setLauncherToAnyAnnulledLauncher(patchAddr, whenAllDoneRestoreToThisInsn);
}

void condBranch_LRAndR::emitStuffAfterPreInsnSnippets(directToMemoryEmitter_t &em,
                                                const function_t &fn,
                                                const instr_t::controlFlowInfo*
                                                ) const {
   const unsigned initialNumInsnBytesEmitted = em.numInsnBytesEmitted();
   kptr_t currInsnAddr = patchAddr + em.numInsnBytesEmitted();
   bool relSuccess = false;
   instr_t *origBranch =    fn.get1OrigInsn(launchSiteAddr)
   instr_t *relocatedBranch = 
     origBranch->tryRelocationNoLink(launchSiteAddr, currInsnAddr, &relSuccess);
   if (relSuccess) {
     em.emit(relocatedBranch);
   } else {
     //double check that we really are out of range
     assert(!power_instr::inRangeOfBranchCond(launchSiteAddr, currInsnAddr));
     assert(cfi.destination == pcRelative); //if absolute, how can it be out of range ?
     //this better work
     assert(power_instr::inRangeOfBranch(launchSiteAddr, currInsnAddr));
     
     //replace a "conditional" branch with a short-range conditional branch + a  long jump
     delete relocatedBranch; //useless now
     
     //the new short-range branch has same condition as the original, but a displacement of 8
     instr_t *shortBranch = new power_instr(power_instr::branchcond, origBranch->getBOField(),
					    origBranch->getBIField(), 8, 0, 0);
     em.emit(shortBranch);

   }
   //jump back to "not-taken" case
   assert(power_instr::inRangeOfBranch(currInsnAddr, launchSiteAddr + 4));
   kaddrdiff_t jump_distance = launchSiteAddr + 4 - currInsnAddr;
   simmeidate<24> li = jump_distance/4; //number of words
   instr_t *theJumpBack = new power_instr(power_instr::branch, li, 0 /*not absolute */, 0 /*no link */ );
   em.emit(theJumpBack);

   if (!relSuccess) {
     //the jump to taken case
     kaddrdiff_t new_displacement = launchSiteAddr + cfi.offset_nbytes - currInsnAddr;
     li = new_displacement/4; //number of words
     relocatedBranch = new power_instr(power_instr::branch, li, 0/*not absolute */ , 0 /*no linking */); 
     em.emit(relocatedBranch);
   }


   assert(em.numInsnBytesEmitted() - initialNumInsnBytesEmitted <= conservative_numInsnBytesForRAndR());
}

