/* symbol_table.c */

#include <linux/module.h>
#include <linux/slab.h>
#include <kallsyms.h>
#include <asm/uaccess.h>
#include "symbol_table.h"

//extern struct module *module_list;

unsigned kerninst_symtab_do(int fillin, unsigned char *buf)
{
  unsigned nbytes = 0, nmods = 0, name_len = 0, nsyms = 0;
  kptr_t data_addr = 0, text_addr = 0;
  unsigned data_size = 0, text_size = 0;
  unsigned str_tab_len, needed_tab_len, name_offset = 0;
  int nsec = 0;
  const char *name = NULL;
  const char *mod_name = NULL;
  unsigned char *bufp = buf;
  unsigned int *len_ptr = NULL;
  struct module *start = THIS_MODULE;
  //struct module *start = module_list;
  struct module *modp = start;  
  struct module_symbol *symp = NULL;

  const struct kallsyms_header *ka_hdr = NULL;
  const struct kallsyms_section *ka_sec = NULL;
  const struct kallsyms_symbol *ka_sym = NULL;
  const struct kallsyms_symbol *ka_sym_tmp = NULL;
  struct kallsyms_symbol ksym;
  const char *ka_str = NULL;
 
  /* sanity check */
  if( fillin && !bufp ) return 0;

  /* record number of modules, including kernel */
  do {
    nmods++;
    modp = modp->next;
  } while((modp != NULL) && (modp != start));
  nbytes += sizeof(unsigned); /* to hold number of modules */
  if(fillin) { 
    *(unsigned *)bufp = nmods;
    bufp += sizeof(unsigned); 
  }

  /* record info on each module */
  modp = start;
  nbytes += nmods*sizeof(unsigned); /* to hold module name length for each */

  do {
    symp = modp->syms;
    mod_name = modp->name;
    /* get module name information */
    name_len = strlen(mod_name);
    if(fillin) {
      if(name_len == 0) { /* the "kernel" module */
        *(unsigned *)bufp = 6;
        bufp += sizeof(unsigned);
        strncpy(bufp, "kernel", 6);
        bufp += 6;
      } else { /* regular module */
        *(unsigned *)bufp = name_len;
        bufp += sizeof(unsigned);
        strncpy(bufp, mod_name, name_len);
        bufp += name_len;
      }
    }
    nbytes += ( name_len ? name_len : 6 );

    /* get rest of module information */
    if( mod_member_present(modp,kallsyms_start) && 
        mod_member_present(modp,kallsyms_end) &&
        (modp->kallsyms_start < modp->kallsyms_end) ) { 
      /* the module is giving us what we want -> ALL the symbols */
      ka_hdr = (struct kallsyms_header *)modp->kallsyms_start;
      ka_sec = (struct kallsyms_section *)((char *)ka_hdr + ka_hdr->section_off);
      ka_sym = (struct kallsyms_symbol *)((char *)ka_hdr + ka_hdr->symbol_off);
      ka_str = ((char *)ka_hdr + ka_hdr->string_off);

      nsec = ka_hdr->sections;
      nbytes += 2 * (sizeof(kptr_t) + sizeof(unsigned));
        /* to hold start & size of text & data sections */
      if(fillin) {
        while(nsec--) {
          name = ka_str + ka_sec->name_off;
          if( ! strcmp(name, ".data") ) {
            data_addr = ka_sec->start;
            data_size = ka_sec->size;

          } else if( ! strcmp(name, ".text") ) {
            text_addr = ka_sec->start;
            text_size = ka_sec->size;
          }
          if(nsec)
            kallsyms_next_sec(ka_hdr, ka_sec); 
        }
        *(kptr_t *)bufp = text_addr;
        bufp += sizeof(kptr_t);
        *(unsigned *)bufp = text_size;
        bufp += sizeof(unsigned);
        *(kptr_t *)bufp = data_addr;
        bufp += sizeof(kptr_t);
        *(unsigned *)bufp = data_size;
        bufp += sizeof(unsigned);
      }

      /* loop through all symbols to find size of string table */
      nsyms = ka_hdr->symbols;
      ka_sym_tmp = ka_sym;
      str_tab_len = 0;
      needed_tab_len = 0;
      while(nsyms--) {
	name_len = strlen(ka_str + ka_sym_tmp->name_off);
	needed_tab_len = ka_sym_tmp->name_off + name_len + 1;
	if(needed_tab_len > str_tab_len)
	  str_tab_len = needed_tab_len;
	if(nsyms)
	  kallsyms_next_sym(ka_hdr, ka_sym_tmp);
      }
      nbytes += sizeof(unsigned) + str_tab_len;
        /* to hold string table length & the table itself */
      if(fillin) {
	*(unsigned *)bufp = str_tab_len;
	bufp += sizeof(unsigned);
        //printk(KERN_WARNING "about to call memcpy #1\n");
	memcpy(bufp, ka_str, str_tab_len);
	bufp += str_tab_len;
      }

      /* a cookie marking the beginning of symbols */
      if (fillin) {
         *(unsigned*)bufp = 0x12345678;
	 bufp += sizeof(unsigned);
      }
      nbytes += sizeof(unsigned);

      /* now process symbols */
      nsyms = ka_hdr->symbols;
      nbytes += sizeof(unsigned); /* to hold number of symbols */
      if(fillin) {
        *(unsigned *)bufp = nsyms;
        bufp += sizeof(unsigned);
      }
      nbytes += nsyms * (sizeof(struct kallsyms_symbol));
      if(fillin) {
	while(nsyms--) {
          //printk(KERN_WARNING "about to call memcpy #2\n");
          memcpy(bufp, ka_sym, sizeof(struct kallsyms_symbol));
	  bufp += sizeof(struct kallsyms_symbol);
	  if(nsyms) 
	    kallsyms_next_sym(ka_hdr, ka_sym);
	}
      }

    } else { /* we'll have to settle for the exported symbols */

      nbytes += 2 * (sizeof(kptr_t) + sizeof(unsigned));
        /* to hold start & size of text & data sections */
      if(fillin) {
        *(kptr_t *)bufp = 0;
        bufp += sizeof(kptr_t);
        *(unsigned *)bufp = 0;
        bufp += sizeof(unsigned);
        *(kptr_t *)bufp = 0;
        bufp += sizeof(kptr_t);
        *(unsigned *)bufp = 0;
        bufp += sizeof(unsigned);
      }

      /* loop through all symbols to find size of string table */
      nsyms = modp->nsyms;
      symp = modp->syms;
      str_tab_len = 0;
      len_ptr = (unsigned int*)bufp;
      bufp += sizeof(unsigned);
      while(nsyms--) {
	name_len = strlen(symp->name) + 1;
	str_tab_len += name_len;
	if(fillin) {
	  strncpy(bufp, symp->name, name_len);
	  bufp += name_len;
	}
	if(nsyms)
	  symp++;
      }
      if(fillin) 
	*len_ptr = str_tab_len;
      nbytes += sizeof(unsigned) + str_tab_len;

      /* a cookie marking the beginning of symbols */
      if (fillin) {
         *(unsigned*)bufp = 0x12345678;
	 bufp += sizeof(unsigned);
      }
      nbytes += sizeof(unsigned);

      /* now process symbols */
      nsyms = modp->nsyms;
      nbytes += sizeof(unsigned); /* to hold number of symbols */
      if(fillin) {
        *(unsigned *)bufp = nsyms;
        bufp += sizeof(unsigned);
      }
      nbytes += nsyms * sizeof(struct kallsyms_symbol);
      if(fillin) {
	symp = modp->syms;
	name_offset = 0;
	while(nsyms--) {
	  ksym.section_off = 0;
	  ksym.symbol_addr = symp->value;
	  ksym.name_off = name_offset;
	  //printk(KERN_WARNING "about to call memcpy #3\n");
          memcpy(bufp, &ksym, sizeof(struct kallsyms_symbol));
	  name_len = strlen(symp->name);
	  name_offset += name_len + 1;
          bufp += sizeof(struct kallsyms_symbol);
	  if(nsyms)
	    symp++;
        }
	if(name_offset != str_tab_len)
	  printk(KERN_WARNING "mismatch on string table\n");
      }
    }
    modp = modp->next;
  } while((modp != NULL) && (modp != start));

  return nbytes;
}

unsigned kerninst_symtab_size()
{
  return kerninst_symtab_do(0, NULL);
}

unsigned kerninst_symtab_get(kptr_t user_buf, int buf_size)
{
  void *buffer;
  unsigned result;
  u_long pages, order = 0, base = 1;
  if( buf_size < kerninst_symtab_size() ) {
    printk(KERN_WARNING "kerninst_symtab_get: buffer too small for symbol table info\n");
    return 0;
  }
  pages = (buf_size / 4096) + ( buf_size % 4096 ? 1 : 0 );
  while( base < pages ) {
    base <<= 1;
    order++;
  }
  buffer = (void *)__get_free_pages(GFP_KERNEL, order);
  if(buffer) {
    result = kerninst_symtab_do(1, (unsigned char *)buffer);
    copy_to_user((void *)user_buf, buffer, result);
    free_pages((u_long)buffer, order);
    return result;
  }
  return 0;
}
